# -*- coding: utf-8 -*-
r"""Module implementing spatial feature recognition (localization) in Consenrich-estimated genomic signals. Applies a simple matched-filtering approach using discrete wavelet-based templates and resampling strategy to determine significance.
"""

import logging
from typing import List, Optional

import pandas as pd
import pywt as pw
import numpy as np
import numpy.typing as npt

from scipy import signal, stats

from . import cconsenrich

logging.basicConfig(level=logging.INFO,
                     format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logging.basicConfig(level=logging.WARNING,
                    format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


def matchWavelet(chromosome: str,
    intervals: npt.NDArray[np.float64],
    values: npt.NDArray[np.float64],
    templateNames: List[str],
    cascadeLevels: List[int],
    iters: int,
    alpha: float,
    minMatchLengthBP: Optional[int],
    maxNumMatches: Optional[int] = 10_000,
    minSignalAtMaxima: Optional[float] = None,
    randSeed: int = 42
    ) -> pd.DataFrame:
    r"""(Experimental) Identify genomic regions showing **enrichment** and **relevant, non-random structure** reinforced in multiple samples.

    :param chromosome: Chromosome name (e.g., 'chr1').
    :type chromosome: str
    :param intervals: Index set of genomic intervals (e.g., [0, 50, 100, ...] nucleotides).
    :type intervals: npt.NDArray[np.float64]
    :param values: Multi-sample signal estimates, e.g., from Consenrich.
    :type values: npt.NDArray[np.float64]
    :param templateNames: List of wavelet template names to use for matching.
    :type templateNames: List[str]
    :param cascadeLevels: List of 'levels' 
    :type cascadeLevels: List[int]
    :param iters: Number of random blocks to sample while building the empirical distribution.
    :type iters: int
    :param alpha: Significance level for thresholding the response sequence.
    :type alpha: float

    :seealso: :func:`cconsenrich.csampleBlockStats`, :class:`consenrich.core.matchingParams`
    """
    randSeed_: int = int(randSeed)
    intervalLengthBP = intervals[1] - intervals[0]
    cols = ['chromosome', 'start', 'end', 'name', 'score', 'strand', 'signal', 'pValue', 'qValue', 'pointSource']
    matchDF = pd.DataFrame(columns=cols)

    for templateName, cascadeLevel in zip(templateNames, cascadeLevels):
        try:
            templateName = str(templateName)
            cascadeLevel = int(cascadeLevel)
        except ValueError:
            logger.warning(f"Skipping invalid templateName or cascadeLevel: {templateName}, {cascadeLevel}")
            continue
        if templateName not in pw.wavelist(kind='discrete'):
            logger.warning(f"\nSkipping unknown wavelet template: {templateName}\nAvailable templates: {pw.wavelist(kind='discrete')}")
            continue

        wav = pw.Wavelet(templateName)
        scalingFunc, waveletFunc, x = wav.wavefun(level=cascadeLevel)
        template = np.array(waveletFunc, dtype=np.float64)/np.linalg.norm(waveletFunc)
        responseSequence: npt.NDArray[np.float64] = signal.fftconvolve(values, template, mode='same')

        if minMatchLengthBP is None:
            minMatchLengthBP = len(template)*intervalLengthBP*2
        relativeMaximaWindow = int(minMatchLengthBP // intervalLengthBP)

        if relativeMaximaWindow % 2 == 0 or relativeMaximaWindow < 1:
            relativeMaximaWindow += 1

        logger.info(f"\nSampling {iters} block maxima for template {templateName} at cascade level {cascadeLevel} with relative maxima window size {relativeMaximaWindow}.")
        blockMaxima = cconsenrich.csampleBlockStats(responseSequence, relativeMaximaWindow, iters, randSeed_)
        responseThreshold = np.quantile(blockMaxima, 1-alpha)
        ecdfBlockMaximaSF = stats.ecdf(blockMaxima).sf
        logger.info(f"Done. Sampled {len(blockMaxima)} block maxima to determine threshold: {responseThreshold:.3f}.\n")

        signalThreshold: float = 0.0
        if minSignalAtMaxima is not None:
            signalThreshold = np.quantile(values, 0.75)

        relativeMaximaIndices = signal.argrelmax(responseSequence, order=relativeMaximaWindow)[0]
        relativeMaximaIndices = relativeMaximaIndices[(responseSequence[relativeMaximaIndices] > responseThreshold) & (values[relativeMaximaIndices] > signalThreshold)]

        if maxNumMatches is not None:
            if len(relativeMaximaIndices) > maxNumMatches:
                relativeMaximaIndices = relativeMaximaIndices[np.argsort(responseSequence[relativeMaximaIndices])[-maxNumMatches:]]

        if len(relativeMaximaIndices) == 0:
            logger.warning(f"no matches were detected using for template {templateName} at cascade level {cascadeLevel}.")
            continue

        starts = intervals[relativeMaximaIndices] - minMatchLengthBP
        ends = intervals[relativeMaximaIndices] + minMatchLengthBP
        pointSources = ((starts + ends) // 2) - starts
        minResponse = np.min(responseSequence[relativeMaximaIndices])
        if minResponse < 0:
            responseSequence[relativeMaximaIndices] -= minResponse
            minResponse = 0.0
        maxResponse = np.max(responseSequence[relativeMaximaIndices])
        rangeResponse = max(maxResponse - minResponse, 1.0)
        scores = 250 + 750*(responseSequence[relativeMaximaIndices] - minResponse) / rangeResponse
        names = [f'{templateName}_{cascadeLevel}_{i}' for i in relativeMaximaIndices]
        strands = ['.' for _ in range(len(scores))]
        pValues = (-np.log10(np.clip(ecdfBlockMaximaSF.evaluate(responseSequence[relativeMaximaIndices]), 1e-8, 1.0))).astype(np.float32)
        qValues = np.array(np.ones_like(pValues)*-1, dtype=np.float32)

        tempDF = pd.DataFrame({
            'chromosome': [chromosome]*len(relativeMaximaIndices),
            'start': starts.astype(int),
            'end': ends.astype(int),
            'name': names,
            'score': scores,
            'strand': strands,
            'signal': responseSequence[relativeMaximaIndices],
            'pValue': pValues,
            'qValue': qValues,
            'pointSource': pointSources.astype(int)
            })
        if matchDF.empty:
            matchDF = tempDF
        else:
            matchDF = pd.concat([matchDF, tempDF], ignore_index=True)
        randSeed_ += 1

    if matchDF.empty:
        logger.warning(f"No matches detected, returning empty DataFrame.")
        return matchDF
    matchDF.sort_values(by=['chromosome', 'start', 'end'], inplace=True)
    matchDF.reset_index(drop=True, inplace=True)
    return matchDF





# -*- coding: utf-8 -*-
# cython: boundscheck=False, wraparound=False, cdivision=True, language_level=3
r"""Cython module for Consenrich core functions.

This module contains Cython implementations of core functions used in Consenrich.
"""

from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int64_t
from libc.math cimport sqrt, fabs
import numpy as np
cimport numpy as cnp
from pysam.libcalignmentfile cimport AlignmentFile, AlignedSegment
cimport cython

cnp.import_array()

cpdef int stepAdjustment(int value, int stepSize, int pushForward=0):
    r"""Adjusts a value to the nearest multiple of stepSize, optionally pushing it forward.

    :param value: The value to adjust.
    :type value: int
    :param stepSize: The step size to adjust to.
    :type stepSize: int
    :param pushForward: If non-zero, pushes the value forward by stepSize if it is
        not already a multiple of stepSize.
    :type pushForward: int
    :return: The adjusted value.
    :rtype: int
    """
    return max(0, (value - (value % stepSize))) + pushForward*stepSize


cpdef uint64_t cgetFirstChromRead(str bamFile, str chromosome, uint64_t chromLength, uint32_t samThreads, int samFlagExclude):
    r"""Get the start position of the first read in a BAM file for a given chromosome.

    :param bamFile: See :func:`consenrich.core.inputParams`.
    :type bamFile: str
    :param chromosome: Chromosome name.
    :type chromosome: str
    :param chromLength: Length of the chromosome in base pairs.
    :type chromLength: uint64_t
    :param samThreads: Number of threads to use for reading the BAM file.
    :type samThreads: uint32_t
    :param samFlagExclude: SAM flags to exclude reads (e.g., unmapped,
    :type samFlagExclude: int
    :return: Start position of the first read in the chromosome, or 0 if no reads are found.
    :rtype: uint64_t
    """

    cdef AlignmentFile aln = AlignmentFile(bamFile, 'rb', threads=samThreads)
    cdef AlignedSegment read
    for read in aln.fetch(contig=chromosome, start=0, stop=chromLength):
        if not (read.flag & samFlagExclude):
            aln.close()
            return read.reference_start
    aln.close()
    return 0


cpdef uint64_t cgetLastChromRead(str bamFile, str chromosome, uint64_t chromLength, uint32_t samThreads, int samFlagExclude):
    r"""Get the end position of the last read in a BAM file for a given chromosome.

    :param bamFile: See :func:`consenrich.core.inputParams`.
    :type bamFile: str
    :param chromosome: Chromosome name.
    :type chromosome: str
    :param chromLength: Length of the chromosome in base pairs.
    :type chromLength: uint64_t
    :param samThreads: Number of threads to use for reading the BAM file.
    :type samThreads: uint32_t
    :param samFlagExclude: See :class:`consenrich.core.samParams`.
    :type samFlagExclude: int
    :return: End position of the last read in the chromosome, or 0 if no reads are found.
    :rtype: uint64_t
    """

    cdef uint64_t start_ = chromLength - min((chromLength // 2), 1_000_000)
    cdef uint64_t lastPos = 0
    cdef AlignmentFile aln = AlignmentFile(bamFile, 'rb', threads=samThreads)
    cdef AlignedSegment read
    for read in aln.fetch(contig=chromosome, start=start_, end=chromLength):
        if not (read.flag & samFlagExclude):
            lastPos = read.reference_end
    aln.close()
    return lastPos



cpdef uint32_t cgetReadLength(str bamFile, uint32_t minReads, uint32_t samThreads, uint32_t maxIterations, int samFlagExclude):
    r"""Get the median read length from a BAM file after fetching a specified number of reads.

    :param bamFile: see :class:`consenrich.core.inputParams`.
    :type bamFile: str
    :param minReads: Minimum number of reads to consider for the median calculation.
    :type minReads: uint32_t
    :param samThreads: See :class:`consenrich.core.samParams`.
    :type samThreads: uint32_t
    :param maxIterations: Maximum number of reads to iterate over.
    :type maxIterations: uint32_t
    :param samFlagExclude: See :class:`consenrich.core.samParams`.
    :type samFlagExclude: int
    :return: Median read length from the BAM file.
    :rtype: uint32_t
    """
    cdef uint32_t observedReads = 0
    cdef uint32_t currentIterations = 0
    cdef AlignmentFile aln = AlignmentFile(bamFile, 'rb', threads=samThreads)
    cdef AlignedSegment read
    cdef cnp.ndarray[cnp.uint32_t, ndim=1] readLengths = np.zeros(maxIterations, dtype=np.uint32)
    cdef uint32_t i = 0
    if <uint32_t>aln.mapped < minReads:
        aln.close()
        return 0
    for read in aln.fetch():
        if not (observedReads < minReads and currentIterations < maxIterations):
            break
        if not (read.flag & samFlagExclude):
            # meets critera -> add it
            readLengths[i] = read.query_length
            observedReads += 1
            i += 1
        currentIterations += 1
    aln.close()
    if observedReads < minReads:
        return 0

    return <uint32_t>np.median(readLengths[:observedReads])

cpdef cnp.uint32_t[:] creadBamSegment(
    str bamFile,
    str chromosome,
    uint32_t start,
    uint32_t end,
    uint32_t stepSize,
    uint32_t readLength,
    uint8_t oneReadPerBin,
    uint16_t samThreads,
    uint16_t samFlagExclude,
    str offsetStr):
    r"""Count reads in a BAM file for a given chromosome"""

    cdef Py_ssize_t nBins = ((end - start + stepSize) // stepSize)
    cdef cnp.ndarray[cnp.uint32_t, ndim=1] values_np = np.zeros(nBins, dtype=np.uint32)
    cdef cnp.uint32_t[::1] values = values_np
    cdef AlignmentFile aln = AlignmentFile(bamFile, 'rb', threads=samThreads)
    cdef AlignedSegment read
    cdef int64_t start64 = start
    cdef int64_t step64 = stepSize
    cdef Py_ssize_t i, index0, index1, midIndex
    cdef Py_ssize_t lastIndex = nBins - 1
    cdef int64_t fwdLeft = 0
    cdef int64_t fwdRight = 0
    cdef int64_t revLeft = 0
    cdef int64_t revRight = 0

    cdef object parts = [p.strip() for p in offsetStr.split(',') if p.strip()]
    if len(parts) == 2 or len(parts) == 0:
        fwdLeft = revLeft = <int64_t> int(parts[0])
        fwdRight = revRight = <int64_t> int(parts[1])
    elif len(parts) == 4:
        fwdLeft  = <int64_t> int(parts[0])
        fwdRight = <int64_t> int(parts[1])
        revLeft  = <int64_t> int(parts[2])
        revRight = <int64_t> int(parts[3])

    cdef uint16_t flag
    cdef bint isReverse
    cdef int64_t rs, re
    cdef int64_t adjStart, adjEnd, midPoint
    cdef int64_t offsetStart, offsetEnd

    try:
        if oneReadPerBin == 0 and readLength > stepSize:
            for read in aln.fetch(chromosome, start, end):
                flag = <uint16_t> read.flag
                if (flag & samFlagExclude) != 0:
                    continue

                isReverse = (flag & 16) != 0
                rs = <int64_t> read.reference_start
                re = <int64_t> read.reference_end

                if not isReverse:
                    adjStart = rs + fwdLeft
                    adjEnd   = re + fwdRight
                else:
                    adjStart = rs + revRight
                    adjEnd   = re - revLeft

                if adjStart > adjEnd:
                    continue

                offsetStart = adjStart - start64
                offsetEnd   = adjEnd - start64

                index0 = <Py_ssize_t> (offsetStart // step64)
                index1 = <Py_ssize_t> (offsetEnd   // step64)

                if index0 < 0:
                    index0 = 0
                if index1 > lastIndex:
                    index1 = lastIndex

                if index0 <= index1:
                    for i in range(index0, index1 + 1):
                        values[i] += 1
        else:
            for read in aln.fetch(chromosome, start, end):
                flag = <uint16_t> read.flag
                if (flag & samFlagExclude) != 0:
                    continue

                isReverse = (flag & 16) != 0
                rs = <int64_t> read.reference_start
                re = <int64_t> read.reference_end

                if not isReverse:
                    adjStart = rs + fwdLeft
                    adjEnd   = re + fwdRight
                else:
                    adjStart = rs + revRight
                    adjEnd   = re - revLeft

                if adjStart > adjEnd:
                    continue

                midPoint = (adjStart + adjEnd) >> 1
                midIndex = <Py_ssize_t> ((midPoint - start64) // step64)
                if 0 <= midIndex < nBins:
                    values[midIndex] += 1
    finally:
        aln.close()

    return values


cpdef cnp.ndarray[cnp.float64_t, ndim=2] cinvertMatrixE(cnp.float64_t[:] muncMatrixIter, double priorCovarianceOO):
    r"""Invert the residual covariance matrix during the forward pass.

    :param muncMatrixIter: The diagonal elements of the covariance matrix at a given genomic interval.
    :type muncMatrixIter: cnp.ndarray[cnp.float64_t, ndim=1]
    :param priorCovarianceOO: The a priori 'primary' state variance :math:`P_{[i|i-1,11]}`.
    :type priorCovarianceOO: cnp.float64_t
    :return: The inverted covariance matrix.
    :rtype: cnp.ndarray[cnp.float64_t, ndim=2]
    """

    cdef int m = muncMatrixIter.size
    # we have to invert a P.D. covariance (diagonal) and rank-one (1*priorCovariance) matrix
    cdef cnp.ndarray[cnp.float64_t, ndim=2] inverse = np.empty((m, m), dtype=np.float64)
    # note, not actually an m-dim matrix, just the diagonal elements taken as input
    cdef cnp.ndarray[cnp.float64_t, ndim=1] muncMatrixInverse = np.empty(m, dtype=np.float64)
    cdef double sqrtPrior = sqrt(priorCovarianceOO)
    cdef cnp.ndarray[cnp.float64_t, ndim=1] uVec = np.empty(m, dtype=np.float64)
    cdef double divisor = 1.0
    cdef double scale
    cdef double uVecI
    cdef Py_ssize_t i, j
    for i in range(m):
        # two birds: build up the trace while taking the reciprocals
        muncMatrixInverse[i] = 1.0/(muncMatrixIter[i])
        divisor += priorCovarianceOO*muncMatrixInverse[i]
    # we can combine these two loops, keeping construction
    # of muncMatrixInverse and uVec separate for now in case
    # we want to parallelize this later
    for i in range(m):
        uVec[i] = sqrtPrior*muncMatrixInverse[i]
    scale = 1.0 / divisor
    for i in range(m):
        uVecI = uVec[i]
        inverse[i, i] = muncMatrixInverse[i] - (scale*uVecI*uVecI)
        for j in range(i + 1, m):
            inverse[i, j] = -scale * uVecI * uVec[j]
            inverse[j, i] = inverse[i, j]
    return inverse


cpdef cnp.ndarray[cnp.float64_t, ndim=1] cgetStateCovarTrace(cnp.ndarray[cnp.float64_t, ndim=3] stateCovarMatrices):
    cdef int n = stateCovarMatrices.shape[0]
    cdef cnp.ndarray[cnp.float64_t, ndim=1] trace = np.empty(n, dtype=np.float64)
    for i in range(n):
        trace[i] = stateCovarMatrices[i, 0, 0] + stateCovarMatrices[i, 1, 1]
    return trace


cpdef cgetPrecisionWeightedResidual(cnp.ndarray[cnp.float64_t, ndim=2] postFitResiduals,
                                    cnp.ndarray[cnp.float64_t, ndim=2] matrixMunc):

    cdef int n = postFitResiduals.shape[0]
    cdef int m = postFitResiduals.shape[1]
    cdef cnp.ndarray[cnp.float64_t, ndim=1] precisionWeightedResidual = np.empty(n, dtype=np.float64)
    cdef cnp.ndarray[cnp.float64_t, ndim=1] weightsIter = np.empty(m, dtype=np.float64)
    cdef cnp.ndarray[cnp.float64_t, ndim=1] residualsIter = np.empty(m, dtype=np.float64)
    cdef double sumWeights = 0.0
    cdef double sumResiduals = 0.0
    cdef int i, j
    for i in range(n):
        sumWeights = 0.0
        sumResiduals = 0.0
        for j in range(m):
            weightsIter[j] = 1.0 / matrixMunc[j, i]
            residualsIter[j] = postFitResiduals[i, j]
            sumWeights += weightsIter[j]
            sumResiduals += residualsIter[j] * weightsIter[j]
        if sumWeights > 0.0:
            precisionWeightedResidual[i] = sumResiduals / sumWeights
        else:
            precisionWeightedResidual[i] = 0.00
    return precisionWeightedResidual


cpdef tuple updateProcessNoiseCovariance(cnp.ndarray[cnp.float64_t, ndim=2] matrixQ,
        cnp.ndarray[cnp.float64_t, ndim=2] matrixQCopy,
        double dStat,
        double dStatAlpha,
        double dStatd,
        double dStatPC,
        bint inflatedQ,
        double maxQ,
        double minQ):
    r"""Adjust process noise covariance matrix :math:`\mathbf{Q}_{[i]}`

    :param matrixQ: Current process noise covariance
    :param matrixQCopy: A copy of the initial original covariance matrix :math:`\mathbf{Q}_{[.]}`
    :param inflatedQ: Flag indicating if the process noise covariance is inflated
    :return: Updated process noise covariance matrix and inflated flag
    :rtype: tuple
    """

    cdef double scaleQ, fac
    if dStat > dStatAlpha:
        scaleQ = sqrt(dStatd * fabs(dStat - dStatAlpha) + dStatPC)
        if matrixQ[0, 0] * scaleQ <= maxQ:
            matrixQ[0, 0] *= scaleQ
            matrixQ[0, 1] *= scaleQ
            matrixQ[1, 0] *= scaleQ
            matrixQ[1, 1] *= scaleQ
        else:
            fac = maxQ / matrixQCopy[0, 0]
            matrixQ[0, 0] = maxQ
            matrixQ[0, 1] = matrixQCopy[0, 1] * fac
            matrixQ[1, 0] = matrixQCopy[1, 0] * fac
            matrixQ[1, 1] = maxQ
        inflatedQ = True

    elif dStat < dStatAlpha and inflatedQ:
        scaleQ = sqrt(dStatd * fabs(dStat - dStatAlpha) + dStatPC)
        if matrixQ[0, 0] / scaleQ >= minQ:
            matrixQ[0, 0] /= scaleQ
            matrixQ[0, 1] /= scaleQ
            matrixQ[1, 0] /= scaleQ
            matrixQ[1, 1] /= scaleQ
        else:
            # we've hit the minimum, no longer 'inflated'
            fac = minQ / matrixQCopy[0, 0]
            matrixQ[0, 0] = minQ
            matrixQ[0, 1] = matrixQCopy[0, 1] * fac
            matrixQ[1, 0] = matrixQCopy[1, 0] * fac
            matrixQ[1, 1] = minQ
            inflatedQ = False
    return np.asarray(matrixQ), inflatedQ


cdef void _blockMax(double[::1] valuesView,
                    Py_ssize_t[::1] blockStartIndices,
                    Py_ssize_t[::1] blockSizes,
                    double[::1] outputView) noexcept:
    cdef Py_ssize_t iterIndex, elementIndex, startIndex, blockLength
    cdef double currentMax, currentValue
    for iterIndex in range(outputView.shape[0]):
        startIndex = blockStartIndices[iterIndex]
        blockLength = blockSizes[iterIndex] # note, length of blocks affects upcoming loop
        currentMax = valuesView[startIndex]
        for elementIndex in range(1, blockLength):
            currentValue = valuesView[startIndex + elementIndex]
            if currentValue > currentMax:
                currentMax = currentValue
        outputView[iterIndex] = currentMax


cpdef csampleBlockStats(cnp.ndarray[cnp.float64_t, ndim=1] values,
                        int expectedBlockSize,
                        int iters,
                        int randSeed):
    r"""Sample contiguous blocks in the response sequence, record maxima, and repeat.

    Used to build an empirical null distribution and determine significance of response outputs. See :func:`consenrich.matching.matchWavelet`

    :param values: The response sequence to sample from.
    :type values: cnp.ndarray[cnp.float64_t, ndim=1]
    :param expectedBlockSize: The expected size (geometric) of the blocks to sample.
    :type expectedBlockSize: int
    :param iters: The number of blocks to sample.
    :type iters: int
    :param randSeed: Random seed for reproducibility.
    :type randSeed: int
    :return: An array of sampled block maxima.
    :rtype: cnp.ndarray[cnp.float64_t, ndim=1]
    """
    cdef cnp.ndarray[cnp.float64_t, ndim=1] valuesArr = np.ascontiguousarray(values, dtype=np.float64)
    cdef double[::1] valuesView = valuesArr
    cdef cnp.ndarray[cnp.intp_t, ndim=1] sizesArr
    cdef cnp.ndarray[cnp.intp_t, ndim=1] startsArr
    cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.empty(iters, dtype=np.float64)
    cdef Py_ssize_t maxBlockLength, maxSize, n
    cdef double maxBlockScale = 10.0
    n = valuesView.shape[0]
    np.random.seed(randSeed)
    sizesArr = np.random.geometric(1.0 / expectedBlockSize, size=iters).astype(np.intp, copy=False)
    maxSize = <int>(maxBlockScale * expectedBlockSize)
    np.clip(sizesArr, 1, maxSize if maxSize < n else n, out=sizesArr)
    maxBlockLength = sizesArr.max() # Py_ssize_t <-- intp ok.
    # by construction, shouldn't exceed the length of the response seq.
    # +1 to check case randint(0,0)
    startsArr = np.random.randint(0, int(n - maxBlockLength + 1), size=iters, dtype=np.intp)

    cdef Py_ssize_t[::1] blockStartIndices = startsArr
    cdef Py_ssize_t[::1] blockSizes = sizesArr
    cdef double[::1] outputView = out
    _blockMax(valuesView, blockStartIndices, blockSizes, outputView)
    return out


def cSparseAvg(double[::1] trackALV, dict sparseMap):
    r"""Fast access and average of `numNearest` sparse elements.

    See :func:`consenrich.core.getMuncTrack`

    :param trackALV: See :func:`consenrich.core.getAverageLocalVarianceTrack`
    :type trackALV: double[::1]
    :param sparseMap: See :func:`consenrich.core.getSparseMap`
    :type sparseMap: dict[int, np.ndarray]
    :return: array of mena('nearest local variances') same length as `trackALV`
    :rtype: cnp.ndarray[cnp.float64_t, ndim=1]
    """
    cdef Py_ssize_t n = <Py_ssize_t>trackALV.shape[0]
    cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.empty(n, dtype=np.float64)
    cdef Py_ssize_t i, j, m
    cdef double sumNearestVariances = 0.0
    cdef cnp.ndarray[cnp.intp_t, ndim=1] idxs
    cdef cnp.intp_t[::1] idx_view
    for i in range(n):
        idxs = <cnp.ndarray[cnp.intp_t, ndim=1]> sparseMap[i] # FFR: to avoid the cast, create sparseMap as dict[intp, np.ndarray[intp]]
        idx_view = idxs
        m = idx_view.shape[0] # FFR: maybe enforce strict `m == numNearest` in future releases to avoid extra overhead
        if m == 0:
            # this case probably warrants an exception or np.nan
            out[i] = 0.0
            continue
        sumNearestVariances = 0.0
        with nogil:
            for j in range(m):
                sumNearestVariances += trackALV[idx_view[j]]
        out[i] = sumNearestVariances/m

    return out


# -*- coding: utf-8 -*-
r"""Various constants and genome resources used in Consenrich.
"""

import logging
import os
from typing import List, Optional

logging.basicConfig(level=logging.INFO,
                     format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logging.basicConfig(level=logging.WARNING,
                    format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

EFFECTIVE_GENOME_SIZES = {
    "hg19": {
        50: 2685511454,
        75: 2736124898,
        100: 2776919708,
        150: 2827436883,
        200: 2855463800,
        250: 2855044784,
    },
    "hg38": {
        50: 2701495711,
        75: 2747877702,
        100: 2805636231,
        150: 2862010428,
        200: 2887553103,
        250: 2898802627,
    },
    "t2t": {
        50: 2725240337,
        75: 2786136059,
        100: 2814334875,
        150: 2931551487,
        200: 2936403235,
        250: 2960856300,
    },
    "mm10": {
        50: 2308125299,
        75: 2407883243,
        100: 2467481008,
        150: 2494787038,
        200: 2520868989,
        250: 2538590322,
    },
    "mm39": {
        50: 2309746861,
        75: 2410055689,
        100: 2468088461,
        150: 2495461690,
        200: 2521902382,
        250: 2538633971,
    },
    "dm3": {
        50: 130428510,
        75: 135004387,
        100: 139647132,
        150: 144307658,
        200: 148523810,
        250: 151901455,
    },
    "dm6": {
        50: 125464678,
        75: 127324557,
        100: 129789773,
        150: 129940985,
        200: 132508963,
        250: 132900923,
    },
    "ce11": {
        50: 95159402,
        75: 96945370,
        100: 98259898,
        150: 98721103,
        200: 98672558,
        250: 101271756,
    }
}

def resolveGenomeName(genome: str) -> str:
    r"""Standardize the genome name for consistency
    :param genome: Name of the genome. See :class:`consenrich.core.genomeParams`.
    :type genome: str
    :return: Standardized genome name.
    :rtype: str
    :raises ValueError: If the genome is not recognized.
    """
    genome_ = genome.lower()
    if genome_ in ['hg19', 'grch37']:
        return 'hg19'
    elif genome_ in ['hg38', 'grch38']:
        return 'hg38'
    elif genome_ in ['t2t', 'chm13', 't2t-chm13']:
        return 't2t'
    elif genome_ in ['mm10', 'grcm38']:
        return 'mm10'
    elif genome_ in ['mm39', 'grcm39']:
        return 'mm39'
    elif genome_ in ['dm3']:
        return 'dm3'
    elif genome_ in ['dm6']:
        return 'dm6'
    elif genome_ in ['ce10','ws220']:
        return 'ce10'
    elif genome_ in ['ce11', 'wbcel235']:
        return 'ce11'
    raise ValueError(f"Genome {genome} is not recognized. Please provide a valid genome name or manually specify resources")


def getEffectiveGenomeSize(genome: str, readLength: int) -> int:
    r"""Get the effective genome size for a given genome and read length.

    :param genome: Name of the genome. See :func:`consenrich.constants.resolveGenomeName` and :class:`consenrich.core.genomeParams`.
    :type genome: str
    :param readLength: Length of the reads. See :func:`consenrich.core.getReadLength`.
    :type readLength: int
    :raises ValueError: If the genome is not recognized or if the read length is not available for the genome.
    :return: Effective genome size in base pairs.
    :rtype: int
    """

    global EFFECTIVE_GENOME_SIZES
    genome_: str = resolveGenomeName(genome)
    if genome_ in EFFECTIVE_GENOME_SIZES:
        if readLength not in EFFECTIVE_GENOME_SIZES[genome_]:
            nearestReadLength: int = int(min(EFFECTIVE_GENOME_SIZES[genome_].keys(), key=lambda x: abs(x - readLength)))
            return EFFECTIVE_GENOME_SIZES[genome_][nearestReadLength]
        return EFFECTIVE_GENOME_SIZES[genome_][readLength]
    raise ValueError(f"Defaults not available for {genome}")


def getGenomeResourceFile(genome: str, fileType: str, dir_: str='data'):
    r"""Get the path to a genome resource file.

    :param genome: the genome assembly. See :func:`consenrich.constants.resolveGenomeName` and :class:`consenrich.core.genomeParams`.
    :type genome: str
    :param fileType: One of 'sizes', 'blacklist', 'sparse'.
    :type fileType: str
    :return: Path to the resource file.
    :rtype: str
    :raises ValueError: If not a sizes, blacklist, or sparse file.
    :raises FileNotFoundError: If the resource file does not exist.
    """
    if fileType.lower() in ['sizes']:
        fileName = f"{genome}.sizes"
    elif fileType.lower() in ['blacklist']:
        fileName = f"{genome}_blacklist.bed"
    elif fileType.lower() in ['sparse']:
        fileName = f"{genome}_sparse.bed"
    filePath = os.path.join(os.path.dirname(__file__), os.path.join(dir_, fileName))
    if not os.path.exists(filePath):
        raise FileNotFoundError(f"Resource file {filePath} does not exist.")
    return filePath


# -*- coding: utf-8 -*-
r"""
==============================================================================
`consenrich.misc_util` -- Miscellaneous utility functions
==============================================================================

"""
import os
from typing import List, Optional, Tuple
import logging
import re
import numpy as np
import pandas as pd
import pybedtools as bed
import pysam as sam

from scipy import signal, ndimage

logging.basicConfig(level=logging.INFO,
                     format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logging.basicConfig(level=logging.WARNING,
                    format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)



def checkBamFile(bamFile: str) -> bool:
    r"""Check that the bam file exists and is indexed

    Assumes the bam file is sorted by coordinates
    """
    has_index = False
    if not os.path.exists(bamFile):
        raise FileNotFoundError(f'Could not find {bamFile}')
    try:
        bamfile = sam.AlignmentFile(bamFile, "rb")
        has_index = bamfile.check_index()
        bamfile.close()
    except AttributeError as aex:
        logger.info(f'Alignments must be in BAM format:\n{aex}')
        raise
    except ValueError as vex:
        has_index = False
        pass

    if not has_index:
        try:
            logger.info(f'Could not find index file for {bamFile}. Calling pysam.index()')
            sam.index(bamFile)
            has_index = True
        except Exception as ex:
            logger.warning(f'Encountered the following exception\n{ex}\nCould not create index file for {bamFile}: is it sorted?')

    return has_index


def getChromSizesDict(sizes_file: str,
                        excludeRegex: str=r'^chr[A-Za-z0-9]+$',
                        excludeChroms: Optional[List[str]] = None) -> dict:
    r"""The function getChromSizesDict is a helper to get chromosome sizes file as a dictionary.
    :param sizes_file: Path to a genome assembly's chromosome sizes file
    :param exclude_regex: Regular expression to exclude chromosomes. Default: all non-standard chromosomes.
    :param exclude_chroms: List of chromosomes to exclude.
    :return: Dictionary of chromosome sizes. Formatted as `{chromosome_name: size}`
    """
    if excludeChroms is None:
        excludeChroms = []
    return {k: v for k, v in pd.read_csv(sizes_file, sep='\t', header=None, index_col=0, names=['chrom','size'])['size'].to_dict().items() if re.search(excludeRegex, k) is not None and k not in excludeChroms}





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import glob
import logging
import pprint
import os
from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any, Union
import shutil
import subprocess
import sys
import numpy as np
import pandas as pd
import pysam
import yaml

import consenrich.core as core
import consenrich.misc_util as misc_util
import consenrich.constants as constants
import consenrich.detrorm as detrorm
import consenrich.matching as matching


logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logging.basicConfig(level=logging.WARNING,
                    format='%(asctime)s - %(module)s.%(funcName)s -  %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


def _listOrEmpty(list_):
    if list_ is None:
        return []
    return list_


def _getMinR(cfg, numBams: int) -> float:
    try:
        raw = cfg.get('observationParams.minR', None)
        return float(raw) if raw is not None else (1/numBams) + 1e-4
    except (TypeError, ValueError, KeyError):
        fallBackMinR: float = 1.0e-2
        logger.warning(f"Invalid or missing 'observationParams.minR' in config. Using `{fallBackMinR}`.")
        return fallBackMinR


def checkControlsPresent(inputArgs: core.inputParams) -> bool:
    """Check if control BAM files are present in the input arguments.

    :param inputArgs: core.inputParams object
    :return: True if control BAM files are present, False otherwise.
    """
    return bool(inputArgs.bamFilesControl) and isinstance(inputArgs.bamFilesControl, list) and len(inputArgs.bamFilesControl) > 0


def getReadLengths(inputArgs: core.inputParams, countingArgs: core.countingParams, samArgs: core.samParams) -> List[int]:
    r"""Get read lengths for each BAM file in the input arguments.

    :param inputArgs: core.inputParams object containing BAM file paths.
    :param countingArgs: core.countingParams object containing number of reads.
    :param samArgs: core.samParams object containing SAM thread and flag exclude parameters.
    :return: List of read lengths for each BAM file.
    """
    if not inputArgs.bamFiles:
        raise ValueError("No BAM files provided in the input arguments.")

    if not isinstance(inputArgs.bamFiles, list) or len(inputArgs.bamFiles) == 0:
        raise ValueError("bam files list is empty")

    return [core.getReadLength(bamFile, countingArgs.numReads, 1000, samArgs.samThreads, samArgs.samFlagExclude)  for bamFile in inputArgs.bamFiles]


def checkMatchingEnabled(matchingArgs: core.matchingParams) -> bool:
    matchingEnabled = (matchingArgs.templateNames is not None) and isinstance(matchingArgs.templateNames, list) and len(matchingArgs.templateNames) > 0
    matchingEnabled = matchingEnabled and (matchingArgs.cascadeLevels is not None) and isinstance(matchingArgs.cascadeLevels, list) and len(matchingArgs.cascadeLevels) > 0
    return matchingEnabled


def getEffectiveGenomeSizes(genomeArgs: core.genomeParams, readLengths: List[int]) -> List[int]:
    r"""Get effective genome sizes for the given genome name and read lengths.
    :param genomeArgs: core.genomeParams object
    :param readLengths: List of read lengths for which to get effective genome sizes.
    :return: List of effective genome sizes corresponding to the read lengths.
    """
    genomeName = genomeArgs.genomeName
    if not genomeName or not isinstance(genomeName, str):
        raise ValueError("Genome name must be a non-empty string.")

    if not isinstance(readLengths, list) or len(readLengths) == 0:
        raise ValueError("Read lengths must be a non-empty list. Try calling `getReadLengths` first.")
    return [constants.getEffectiveGenomeSize(genomeName, readLength) for readLength in readLengths]


def getInputArgs(config_path:str) -> core.inputParams:
    def _expandWildCards(bamList) -> List[str]:
        expanded = []
        for entry in bamList:
            if '*' in entry or '?' in entry or '[' in entry:
                matched = glob.glob(entry)
                expanded.extend(matched)
            else:
                expanded.append(entry)
        return expanded
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    bamFilesRaw = config.get('inputParams.bamFiles', [])
    bamFilesControlRaw = config.get('inputParams.bamFilesControl', [])
    bamFiles = _expandWildCards(bamFilesRaw)
    bamFilesControl = _expandWildCards(bamFilesControlRaw)
    if len(bamFiles) == 0:
        raise ValueError("No BAM files provided in the configuration.")
    if len(bamFilesControl) > 0 and len(bamFilesControl) != len(bamFiles) and len(bamFilesControl) != 1:
        raise ValueError("Number of control BAM files must be 0, 1, or the same as number of treatment files")
    if len(bamFilesControl) == 1:
        # If there are multiple bamFiles, but 1 control, control is applied for all treatment files
        logger.info(f"Only one control given: Using {bamFilesControl[0]} for all treatment files.")
        bamFilesControl = bamFilesControl * len(bamFiles)

    if not bamFiles or not isinstance(bamFiles, list) or len(bamFiles) == 0:
        raise ValueError("No BAM files found")

    for i, bamFile in enumerate(bamFiles):
        misc_util.checkBamFile(bamFile)

    if bamFilesControl:
        for i, bamFile in enumerate(bamFilesControl):
            misc_util.checkBamFile(bamFile)

    return core.inputParams(
        bamFiles=bamFiles,
        bamFilesControl=bamFilesControl
    )


def getGenomeArgs(config_path: str) -> core.genomeParams:
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    genomeName = config.get('genomeParams.name', None)
    genome = constants.resolveGenomeName(genomeName)
    chromSizesFile: Optional[str] = None
    blacklistFile: Optional[str] = None
    sparseBedFile: Optional[str] = None
    chromosomes: Optional[List[str]] = None
    excludeChroms: List[str] = config.get('genomeParams.excludeChroms', [])
    excludeForNorm: List[str] = config.get('genomeParams.excludeForNorm', [])
    if genome:
        chromSizesFile = constants.getGenomeResourceFile(genome, 'sizes')
        blacklistFile = constants.getGenomeResourceFile(genome, 'blacklist')
        sparseBedFile = constants.getGenomeResourceFile(genome, 'sparse')
    if config.get('genomeParams.chromSizesFile', None):
        chromSizesFile = config['genomeParams.chromSizesFile']
    if config.get('genomeParams.blacklistFile', None):
        blacklistFile = config['genomeParams.blacklistFile']
    if config.get('genomeParams.sparseBedFile', None):
        sparseBedFile = config['genomeParams.sparseBedFile']
    if not chromSizesFile or not os.path.exists(chromSizesFile):
        raise FileNotFoundError(f"Chromosome sizes file {chromSizesFile} does not exist.")
    if config.get('genomeParams.chromosomes', None):
        chromosomes = config['genomeParams.chromosomes']
    else:
        if chromSizesFile:
            chromosomes = list(pd.read_csv(chromSizesFile, sep='\t', header=None, names=['chrom', 'size'])['chrom'])
        else:
            raise ValueError("No chromosomes provided in the configuration and no chromosome sizes file specified.")
    chromosomes = [chrom.strip() for chrom in chromosomes if chrom.strip()]
    if excludeChroms:
        chromosomes = [chrom for chrom in chromosomes if chrom not in excludeChroms]
    if not chromosomes:
        raise ValueError("No valid chromosomes found after excluding specified chromosomes.")
    return core.genomeParams(
        genomeName=genome,
        chromSizesFile=chromSizesFile,
        blacklistFile=blacklistFile,
        sparseBedFile=sparseBedFile,
        chromosomes=chromosomes,
        excludeChroms=excludeChroms,
        excludeForNorm=excludeForNorm)

def getCountingArgs(config_path: str) -> core.countingParams:
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    stepSize = config.get('countingParams.stepSize', 25)
    scaleDown = config.get('countingParams.scaleDown', True)
    scaleFactors = config.get('countingParams.scaleFactors', None)
    numReads = config.get('countingParams.numReads', 100)
    scaleFactorsControl = config.get('countingParams.scaleFactorsControl', None)
    if scaleFactors is not None and not isinstance(scaleFactors, list):
        raise ValueError("`scaleFactors` should be a list of floats.")
    if scaleFactorsControl is not None and not isinstance(scaleFactorsControl, list):
        raise ValueError("`scaleFactorsControl` should be a list of floats.")
    if scaleFactors is not None and scaleFactorsControl is not None and len(scaleFactors) != len(scaleFactorsControl):
        if len(scaleFactorsControl) == 1:
            scaleFactorsControl = scaleFactorsControl * len(scaleFactors)
        else:
            raise ValueError("control and treatment scale factors: must be equal length or 1 control")
    return core.countingParams(
        stepSize=stepSize,
        scaleDown=scaleDown,
        scaleFactors=scaleFactors,
        scaleFactorsControl=scaleFactorsControl,
        numReads=numReads,
    )

def readConfig(config_path: str) -> Dict[str, Any]:
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)

    inputParams = getInputArgs(config_path)
    genomeParams = getGenomeArgs(config_path)
    countingParams = getCountingArgs(config_path)
    minR_default = _getMinR(config, len(inputParams.bamFiles))
    return {
        'experimentName': config.get('experimentName', 'consenrichExperiment'),
        'genomeArgs': genomeParams,
        'inputArgs': inputParams,
        'countingArgs': countingParams,
        'processArgs': core.processParams(
            deltaF=config.get('processParams.deltaF', 0.5),
            minQ=config.get('processParams.minQ', 0.25),
            maxQ=config.get('processParams.maxQ', 500.0),
            offDiagQ=config.get('processParams.offDiagQ', 0.0),
            dStatAlpha=config.get('processParams.dStatAlpha', 3.0),
            dStatd=config.get('processParams.dStatd', 10.0),
            dStatPC=config.get('processParams.dStatPC', 2.0)
        ),
        'observationArgs': core.observationParams(
            minR=minR_default,
            maxR=config.get('observationParams.maxR', 500.0),
            useALV=config.get('observationParams.useALV', False),
            useConstantNoiseLevel=config.get('observationParams.useConstantNoiseLevel', False),
            noGlobal=config.get('observationParams.noGlobal', False),
            numNearest=config.get('observationParams.numNearest', 25),
            localWeight=config.get('observationParams.localWeight', 0.333),
            globalWeight=config.get('observationParams.globalWeight', 0.667),
            approximationWindowLengthBP=config.get('observationParams.approximationWindowLengthBP', 10000),
            lowPassWindowLengthBP=config.get('observationParams.lowPassWindowLengthBP', 20000),
            returnCenter=config.get('observationParams.returnCenter', True)
        ),
        'stateArgs': core.stateParams(
            stateInit=config.get('stateParams.stateInit', 0.0),
            stateCovarInit=config.get('stateParams.stateCovarInit', 100.0),
            boundState=config.get('stateParams.boundState', True),
            stateLowerBound=config.get('stateParams.stateLowerBound', 0.0),
            stateUpperBound=config.get('stateParams.stateUpperBound', 10000.0)
        ),
        'samArgs': core.samParams(
            samThreads=config.get('samParams.samThreads', 1),
            samFlagExclude=config.get('samParams.samFlagExclude', 3844),
            oneReadPerBin=config.get('samParams.oneReadPerBin', 0),
            chunkSize=config.get('samParams.chunkSize', 1000000),
            offsetStr=config.get('samParams.offsetStr', "0,0")
        ),
        'detrendArgs': core.detrendParams(
            detrendWindowLengthBP=config.get('detrendParams.detrendWindowLengthBP', 10000),
            detrendTrackPercentile=config.get('detrendParams.detrendTrackPercentile', 75.0),
            usePolyFilter=config.get('detrendParams.usePolyFilter', False),
            detrendSavitzkyGolayDegree=config.get('detrendParams.detrendSavitzkyGolayDegree', 2),
            useOrderStatFilter=config.get('detrendParams.useOrderStatFilter', True)
        ),
        'matchingArgs': core.matchingParams(
            templateNames=config.get('matchingParams.templateNames', []),
            cascadeLevels=config.get('matchingParams.cascadeLevels', []),
            iters=config.get('matchingParams.iters', 10_000),
            alpha=config.get('matchingParams.alpha', 0.01),
            minMatchLengthBP=config.get('matchingParams.minMatchLengthBP', None),
            maxNumMatches=config.get('matchingParams.maxNumMatches', 100_000),
            minSignalAtMaxima=config.get('matchingParams.minSignalAtMaxima', None))
    }

def convertBedGraphToBigWig(experimentName, chromSizesFile):
    suffixes = ['state', 'residuals']
    path_ = ""
    warningMessage = ("Could not find UCSC bedGraphToBigWig binary utility."
        "If you need bigWig files instead of the default, human-readable bedGraph files,"
        "you can download the `bedGraphToBigWig` binary from https://hgdownload.soe.ucsc.edu/admin/exe/<operatingSystem, architecture>"
        "OR install via conda (conda install -c bioconda ucsc-bedgraphtobigwig).")

    logger.info("Attempting to generate bigWig files from bedGraph format...")
    try:
        path_ = shutil.which('bedGraphToBigWig')
    except Exception as e:
        logger.warning(f"\n{warningMessage}\n")
        return
    if path_ is None or len(path_) == 0:
        logger.warning(f"\n{warningMessage}\n")
        return
    logger.info(f"Using bedGraphToBigWig from {path_}")

    for suffix in suffixes:
        bedgraph = f'consenrichOutput_{experimentName}_{suffix}.bedGraph'
        if not os.path.exists(bedgraph):
            logger.warning(f"bedGraph file {bedgraph} does not exist. Skipping sorting + bigWig conversion.")
            continue
        if not os.path.exists(chromSizesFile):
            logger.warning(f"{chromSizesFile} does not exist. Skipping sorting + bigWig conversion.")
            continue
        sorted_bedgraph = bedgraph.replace('.bedGraph', '_sorted.bedGraph')
        bigwig = f'{experimentName}_consenrich_{suffix}.bw'
        subprocess.run(['sort', '-k1,1', '-k2,2n', bedgraph, '-o', sorted_bedgraph], check=True)
        subprocess.run([path_, sorted_bedgraph, chromSizesFile, bigwig], check=True)
        if os.path.exists(bigwig) and os.path.getsize(bigwig) > 100:
            os.remove(sorted_bedgraph)
            logger.info(f"Converted {bedgraph} to {bigwig}.")


def main():
    parser = argparse.ArgumentParser(description="Consenrich CLI")
    parser.add_argument('--config', type=str, dest='config', help='Path to a YAML config file with parameters + arguments defined in `consenrich.core`')
    parser.add_argument('--verbose', action='store_true', help='If set, logs config')
    args = parser.parse_args()

    if not args.config:
        logger.info("No config file provided, run with `--config <path_to_config.yaml>`")
        logger.info("See documentation: https://nolan-h-hamilton.github.io/Consenrich/")
        sys.exit(0)

    if not os.path.exists(args.config):
        logger.info(f"Config file {args.config} does not exist.")
        logger.info("See documentation: https://nolan-h-hamilton.github.io/Consenrich/")
        sys.exit(0)

    config = readConfig(args.config)
    experimentName = config['experimentName']
    genomeArgs = config['genomeArgs']
    inputArgs = config['inputArgs']
    countingArgs = config['countingArgs']
    processArgs = config['processArgs']
    observationArgs = config['observationArgs']
    stateArgs = config['stateArgs']
    samArgs = config['samArgs']
    detrendArgs = config['detrendArgs']
    matchingArgs = config['matchingArgs']
    bamFiles = inputArgs.bamFiles
    bamFilesControl = inputArgs.bamFilesControl
    numSamples = len(bamFiles)
    numNearest = observationArgs.numNearest
    stepSize = countingArgs.stepSize
    excludeForNorm = genomeArgs.excludeForNorm
    chromSizes = genomeArgs.chromSizesFile
    scaleDown = countingArgs.scaleDown

    if args.verbose:
        try:
            logger.info("Configuration:\n")
            config_truncated = {k: v for k, v in config.items() if k not in ['inputArgs', 'genomeArgs', 'countingArgs']}
            config_truncated['experimentName'] = experimentName
            config_truncated['inputArgs'] = inputArgs
            config_truncated['genomeArgs'] = genomeArgs
            config_truncated['countingArgs'] = countingArgs
            config_truncated['processArgs'] = processArgs
            config_truncated['observationArgs'] = observationArgs
            config_truncated['stateArgs'] = stateArgs
            config_truncated['samArgs'] = samArgs
            config_truncated['detrendArgs'] = detrendArgs
            pprint.pprint(config_truncated, indent=4)
        except Exception as e:
            logger.warning(f"Failed to print parsed config:\n{e}\n")


    controlsPresent = checkControlsPresent(inputArgs)
    readLengthsBamFiles = getReadLengths(inputArgs, countingArgs, samArgs)
    effectiveGenomeSizes = getEffectiveGenomeSizes(genomeArgs, readLengthsBamFiles)
    matchingEnabled = checkMatchingEnabled(matchingArgs)
    scaleFactors = countingArgs.scaleFactors
    scaleFactorsControl = countingArgs.scaleFactorsControl

    if controlsPresent:
        treatScaleFactors = []
        controlScaleFactors = []
        if scaleFactors is None or scaleFactorsControl is None:
            readLengthsControlBamFiles = [core.getReadLength(bamFile, countingArgs.numReads, 1000, samArgs.samThreads, samArgs.samFlagExclude) for bamFile in bamFilesControl]
            effectiveGenomeSizesControl = [constants.getEffectiveGenomeSize(genomeArgs.genomeName, readLength) for readLength in readLengthsControlBamFiles]
            pairScalingFactors = [
                detrorm.getPairScaleFactors(bamFileA, bamFileB, effectiveGenomeSizeA, effectiveGenomeSizeB,
                                        readLengthA, readLengthB, excludeForNorm,
                                        chromSizes, samArgs.samThreads, scaleDown)
                for bamFileA, bamFileB, effectiveGenomeSizeA, effectiveGenomeSizeB,
                    readLengthA, readLengthB in zip(bamFiles, bamFilesControl,
                                                    effectiveGenomeSizes, effectiveGenomeSizesControl,
                                                    readLengthsBamFiles, readLengthsControlBamFiles)
            ]
            for scaleFactorA, scaleFactorB in pairScalingFactors:
                treatScaleFactors.append(scaleFactorA)
                controlScaleFactors.append(scaleFactorB)
    else:
        treatScaleFactors = scaleFactors
        controlScaleFactors = scaleFactorsControl


    if scaleFactors is None:
        scaleFactors = [
            detrorm.getScaleFactor1x(bamFile, effectiveGenomeSize, readLength, genomeArgs.excludeChroms, genomeArgs.chromSizesFile, samArgs.samThreads)
            for bamFile, effectiveGenomeSize, readLength in zip(bamFiles, effectiveGenomeSizes, readLengthsBamFiles)
        ]
    chromSizesDict = misc_util.getChromSizesDict(genomeArgs.chromSizesFile, excludeChroms=genomeArgs.excludeChroms)
    chromosomes = genomeArgs.chromosomes


    for c_, chromosome in enumerate(chromosomes):
        chromosomeStart, chromosomeEnd = core.getChromRangesJoint(bamFiles, chromosome, chromSizesDict[chromosome], samArgs.samThreads, samArgs.samFlagExclude)
        chromosomeStart = max(0, (chromosomeStart - (chromosomeStart % stepSize)))
        chromosomeEnd = max(0, (chromosomeEnd - (chromosomeEnd % stepSize)))
        intervals = np.arange(chromosomeStart, chromosomeEnd + stepSize, stepSize)

        chromMat: np.ndarray = np.empty((numSamples, len(intervals)), dtype=np.float64)
        if controlsPresent:
            j_: int = 0
            for bamA, bamB in zip(bamFiles, bamFilesControl):
                pairMatrix: np.ndarray = core.readBamSegments(
                    [bamA, bamB], chromosome, chromosomeStart, chromosomeEnd, stepSize,
                    [readLengthsBamFiles[j_], readLengthsControlBamFiles[j_]], [treatScaleFactors[j_], controlScaleFactors[j_]],
                    samArgs.oneReadPerBin, samArgs.samThreads, samArgs.samFlagExclude, offsetStr=samArgs.offsetStr,
                )
                chromMat[j_, :] = scaleFactors[j_]*(pairMatrix[0, :] - pairMatrix[1, :])
                j_ += 1
        else:
            chromMat = core.readBamSegments(
                bamFiles,
                chromosome,
                chromosomeStart,
                chromosomeEnd,
                stepSize,
                readLengthsBamFiles,
                scaleFactors,
                samArgs.oneReadPerBin,
                samArgs.samThreads,
                samArgs.samFlagExclude,
                offsetStr=samArgs.offsetStr,)
        sparseMap = None
        if genomeArgs.sparseBedFile and not observationArgs.useALV:
            logger.info(f"Building sparse mapping for {chromosome}...")
            sparseMap = core.getSparseMap(chromosome, intervals, numNearest, genomeArgs.sparseBedFile)

        muncMat = np.empty_like(chromMat, dtype=np.float64)
        for j in range(numSamples):
            logger.info(f"Muncing {j+1}/{numSamples} for {chromosome}...")
            muncMat[j, :] = core.getMuncTrack(
                chromosome,
                intervals,
                stepSize,
                chromMat[j, :],
                observationArgs.minR,
                observationArgs.maxR,
                observationArgs.useALV,
                observationArgs.useConstantNoiseLevel,
                observationArgs.noGlobal,
                observationArgs.localWeight,
                observationArgs.globalWeight,
                observationArgs.approximationWindowLengthBP,
                observationArgs.lowPassWindowLengthBP,
                observationArgs.returnCenter,
                sparseMap=sparseMap
            )
            chromMat[j, :] = detrorm.detrendTrack(
                chromMat[j, :],
                stepSize,
                detrendArgs.detrendWindowLengthBP,
                detrendArgs.useOrderStatFilter,
                detrendArgs.usePolyFilter,
                detrendArgs.detrendTrackPercentile,
                detrendArgs.detrendSavitzkyGolayDegree
            )
        logger.info(f">>>Running consenrich: {chromosome}<<<")

        x, P, y = core.runConsenrich(chromMat, muncMat,
        processArgs.deltaF,
        processArgs.minQ,
        processArgs.maxQ,
        processArgs.offDiagQ,
        processArgs.dStatAlpha,
        processArgs.dStatd,
        processArgs.dStatPC,
        stateArgs.stateInit,
        stateArgs.stateCovarInit,
        stateArgs.boundState,
        stateArgs.stateLowerBound,
        stateArgs.stateUpperBound,
        samArgs.chunkSize,
        progressIter=50_000)
        logger.info(f"Done.")


        x_ = core.getPrimaryState(x)
        y_ = core.getPrecisionWeightedResidual(y, muncMat)

        df = pd.DataFrame({
            "Chromosome": chromosome,"Start": intervals, "End": intervals + stepSize,
            "State": x_,
            "Res": y_})
        if c_ == 0 and len(chromosomes) > 1:
            for file_ in os.listdir('.'):
                if file_.startswith(f"consenrichOutput_{experimentName}") and (file_.endswith('.bedGraph') or file_.endswith('.narrowPeak')):
                    os.remove(file_)

        for col, suffix in [("State", "state"),
                            ("Res", "residuals")]:
            df[["Chromosome", "Start", "End", col]].to_csv(
                f"consenrichOutput_{experimentName}_{suffix}.bedGraph", sep="\t",
                header=False, index=False, mode="a", float_format="%.3f",
                lineterminator="\n",
            )
        try:
            if matchingEnabled:
                matchingDF = matching.matchWavelet(
                    chromosome,
                    intervals,
                    x_,
                    matchingArgs.templateNames,
                    matchingArgs.cascadeLevels,
                    matchingArgs.iters,
                    matchingArgs.alpha,
                    matchingArgs.minMatchLengthBP,
                    matchingArgs.maxNumMatches,
                    matchingArgs.minSignalAtMaxima,
                )
                if not matchingDF.empty:
                    matchingDF.to_csv(
                        f"consenrichOutput_{experimentName}_matches.narrowPeak",
                        sep="\t", header=False, index=False, mode="a",
                        float_format="%.3f", lineterminator="\n")
        except Exception as e:
            logger.warning(f"Matching routine unsuccessful for {chromosome}...skipping")
            continue
    logger.info("Finished: output in human-readable format")
    convertBedGraphToBigWig(experimentName, genomeArgs.chromSizesFile)
    logger.info("Done.")

if __name__ == "__main__":
    main()



import os
from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
import pysam
import numpy

def get_includes():
    class Includes:
        def __iter__(self):
            import pysam
            import numpy
            return iter(pysam.get_include()+[numpy.get_include()])
        def __getitem__(self, i):
            return list(self)[i]
    return Includes()

extensions = [
    Extension(
        "consenrich.cconsenrich",
        sources=["src/consenrich/cconsenrich.pyx"],
        include_dirs=get_includes(),
        libraries=pysam.get_libraries(),
    )
]

setup(
    name="consenrich",
    version="0.4.0b1",
    packages=find_packages(where="src"),
    include_package_data=True,
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, language_level="3"),
    install_requires=["cython>=3.0", "numpy>=2.1", "pandas>=2.3", "scipy>=1.15",
                      "pysam>=0.23.3", "pybedtools>=0.11.2", "PyYAML>=6.0.2", "PyWavelets>=1.9.0"],
    python_requires=">=3.11",
    zip_safe=False,
)
